//////////////////////////////////////////////////////////////////////////
// Mapper090  PC-JY-??                                                  //
//////////////////////////////////////////////////////////////////////////
void	Mapper090::Reset()
{
	SetPROM_32K_Bank( PROM_8K_SIZE-4, PROM_8K_SIZE-3, PROM_8K_SIZE-2, PROM_8K_SIZE-1 );
	SetVROM_8K_Bank( 0 );

	irq_enable = 0;	// Disable
	irq_counter = 0;
	irq_latch = 0;

	prg_6000 = 0;
	prg_E000 = 0;
	prg_size = 0;
	chr_size = 0;
	mir_mode = 0;
	mir_type = 0;

	for( INT i = 0; i < 4; i++ ) {
		prg_reg[i] = PROM_8K_SIZE-4+i;
		ntl_reg[i] = 0;
		nth_reg[i] = 0;
		chl_reg[i] = i;
		chh_reg[i] = 0;
		chl_reg[i+4] = i+4;
		chh_reg[i+4] = 0;
	}

	if( sw_val )
		sw_val = 0x00;
	else
		sw_val = 0xFF;

//	nes->SetRenderMethod( NES::PRE_ALL_RENDER );
}

BYTE	Mapper090::ReadLow( WORD addr )
{
#if	0
	if( addr == 0x5000 ) {
		return	(BYTE)(mul_val1*mul_val2);
	} else if( addr >= 0x6000 && addr <= 0x7FFF ) {
		return	Mapper::ReadLow( addr );
	}
	return	(BYTE)(addr>>8);
#else
	return	sw_val?0x00:0xFF;
#endif
}

void	Mapper090::WriteLow( WORD addr, BYTE data )
{
#if	0
	if( addr == 0x5000 ) {
		mul_val1 = data;
	} else
	if( addr == 0x5001 ) {
		mul_val2 = data;
	}
#endif
}

void	Mapper090::Write( WORD addr, BYTE data )
{
	switch( addr & 0xF007 ) {
		case	0x8000:
		case	0x8001:
		case	0x8002:
		case	0x8003:
			prg_reg[addr&3] = data;
			SetBank_CPU();
			break;

		case	0x9000:
		case	0x9001:
		case	0x9002:
		case	0x9003:
		case	0x9004:
		case	0x9005:
		case	0x9006:
		case	0x9007:
			chl_reg[addr&7] = data;
			SetBank_PPU();
			break;

		case	0xA000:
		case	0xA001:
		case	0xA002:
		case	0xA003:
		case	0xA004:
		case	0xA005:
		case	0xA006:
		case	0xA007:
			chh_reg[addr&7] = data;
			SetBank_PPU();
			break;

		case	0xB000:
		case	0xB001:
		case	0xB002:
		case	0xB003:
			ntl_reg[addr&3] = data;
			SetBank_VRAM();
			break;

		case	0xB004:
		case	0xB005:
		case	0xB006:
		case	0xB007:
			nth_reg[addr&3] = data;
			SetBank_VRAM();
			break;


		case	0xC002:
			irq_enable = 0;
			break;

		case	0xC003:
		case	0xC004:
			if( !irq_enable ) {
				irq_enable = 0xFF;
				irq_counter = irq_latch;
			}
			break;

		case	0xC005:
			irq_counter = data;
			irq_latch = data;
			break;

		case	0xD000:
			prg_6000 = data & 0x80;
			prg_E000 = data & 0x04;
			prg_size = data & 0x03;
			chr_size = (data & 0x18)>>3;
			mir_mode = data & 0x20;
			SetBank_CPU();
			SetBank_PPU();
			SetBank_VRAM();
			break;

		case	0xD001:
			mir_type = data & 0x03;
			SetBank_VRAM();
			break;

		case	0xD003:
			break;
	}
}

void	Mapper090::HSync( INT scanline )
{
	if( (scanline >= 0 && scanline <= 239) ) {
		if( nes->ppu->IsDispON() ) {
#if	1
			if( irq_counter )
				irq_counter--;

			if( !irq_counter ) {
				if( irq_enable ) {
					nes->cpu->IRQ();
				}
				irq_enable = 0;
			}
#else
			if( --irq_counter == 0 ) {
				if( irq_enable ) {
					nes->cpu->IRQ();
				}
			}
#endif
		}
	}
}

void	Mapper090::SetBank_CPU()
{
	if( prg_size == 0 ) {
		SetPROM_32K_Bank( PROM_8K_SIZE-4, PROM_8K_SIZE-3, PROM_8K_SIZE-2, PROM_8K_SIZE-1 );
	} else
	if( prg_size == 1 ) {
		SetPROM_32K_Bank( prg_reg[1]*2, prg_reg[1]*2+1, PROM_8K_SIZE-2, PROM_8K_SIZE-1 );
	} else
	if( prg_size == 2 ) {
		if( prg_E000 ) {
			SetPROM_32K_Bank( prg_reg[0], prg_reg[1], prg_reg[2], prg_reg[3] );
		} else {
			if( prg_6000 ) {
				SetPROM_8K_Bank( 3, prg_reg[3] );
			}
			SetPROM_32K_Bank( prg_reg[0], prg_reg[1], prg_reg[2], PROM_8K_SIZE-1 );
		}
	} else {
		SetPROM_32K_Bank( prg_reg[3], prg_reg[2], prg_reg[1], prg_reg[0] );
	}
}

void	Mapper090::SetBank_PPU()
{
INT	bank[8];

	for( INT i = 0; i < 8; i++ ) {
		bank[i] = ((INT)chh_reg[i]<<8)|((INT)chl_reg[i]);
	}

	if( chr_size == 0 ) {
		SetVROM_8K_Bank( bank[0] );
	} else
	if( chr_size == 1 ) {
		SetVROM_4K_Bank( 0, bank[0] );
		SetVROM_4K_Bank( 4, bank[4] );
	} else
	if( chr_size == 2 ) {
		SetVROM_2K_Bank( 0, bank[0] );
		SetVROM_2K_Bank( 2, bank[2] );
		SetVROM_2K_Bank( 4, bank[4] );
		SetVROM_2K_Bank( 6, bank[6] );
	} else {
		SetVROM_8K_Bank( bank[0], bank[1], bank[2], bank[3], bank[4], bank[5], bank[6], bank[7] );
	}
}

void	Mapper090::SetBank_VRAM()
{
INT	bank[4];

	for( INT i = 0; i < 4; i++ ) {
		bank[i] = ((INT)nth_reg[i]<<8)|((INT)ntl_reg[i]);
	}

	if( mir_mode ) {
		for( INT i = 0; i < 4; i++ ) {
			if( !nth_reg[i] && (ntl_reg[i] == (BYTE)i) ) {
				mir_mode = 0;
			}
		}

		if( mir_mode ) {
			SetVROM_1K_Bank(  8, bank[0] );
			SetVROM_1K_Bank(  9, bank[1] );
			SetVROM_1K_Bank( 10, bank[2] );
			SetVROM_1K_Bank( 11, bank[3] );
		}
	} else {
		if( mir_type == 0 ) {
			SetVRAM_Mirror( VRAM_VMIRROR );
		} else
		if( mir_type == 1 ) {
			SetVRAM_Mirror( VRAM_HMIRROR );
		} else {
			SetVRAM_Mirror( VRAM_MIRROR4L );
		}
	}
}

void	Mapper090::SaveState( LPBYTE p )
{
INT	i;

	for( i = 0; i < 4; i++ ) {
		p[i] = prg_reg[i];
	}
	for( i = 0; i < 8; i++ ) {
		p[i+ 4] = chh_reg[i];
	}
	for( i = 0; i < 8; i++ ) {
		p[i+12] = chl_reg[i];
	}
	for( i = 0; i < 4; i++ ) {
		p[i+20] = nth_reg[i];
	}
	for( i = 0; i < 4; i++ ) {
		p[i+24] = ntl_reg[i];
	}
	p[28] = irq_enable;
	p[29] = irq_counter;
	p[30] = irq_latch;
	p[31] = prg_6000;
	p[32] = prg_E000;
	p[33] = prg_size;
	p[34] = chr_size;
	p[35] = mir_mode;
	p[36] = mir_type;
	p[37] = mul_val1;
	p[38] = mul_val2;
}

void	Mapper090::LoadState( LPBYTE p )
{
INT	i;

	for( i = 0; i < 4; i++ ) {
		prg_reg[i] = p[i];
	}
	for( i = 0; i < 8; i++ ) {
		chh_reg[i] = p[i+ 4];
	}
	for( i = 0; i < 8; i++ ) {
		chl_reg[i] = p[i+12];
	}
	for( i = 0; i < 4; i++ ) {
		nth_reg[i] = p[i+20];
	}
	for( i = 0; i < 4; i++ ) {
		ntl_reg[i] = p[i+24];
	}
	irq_enable  = p[28];
	irq_counter = p[29];
	irq_latch   = p[30];
	prg_6000    = p[31];
	prg_E000    = p[32];
	prg_size    = p[33];
	chr_size    = p[34];
	mir_mode    = p[35];
	mir_type    = p[36];
	mul_val1    = p[37];
	mul_val2    = p[38];
}
